@@ -1,35 +1,37 @@
-import csv
 import datetime
 
+import unicodecsv as csv
+
 from django.core.exceptions import ValidationError
 from django.utils.text import slugify
-from django.http import HttpResponse
+from django.http import HttpResponse, StreamingHttpResponse
 
-from django.conf import settings
-if not settings.configured:
-    # required to import ValuesQuerySet
-    settings.configure()  # pragma: no cover
-
-from django.db.models.query import ValuesQuerySet
-
-from django.utils import six
+import six
 
 """ A simple python package for turning django models into csvs """
 
 # Keyword arguments that will be used by this module
 # the rest will be passed along to the csv writer
-DJQSCSV_KWARGS = {'field_header_map': None,
-                  'field_serializer_map': None,
-                  'use_verbose_names': True,
-                  'field_order': None}
+DJQSCSV_KWARGS = {
+    'field_header_map', 'field_serializer_map', 'use_verbose_names',
+    'field_order'}
 
 
 class CSVException(Exception):
     pass
 
 
+class _Echo(object):
+    """An object that implements just the write method of the file-like
+    interface.
+    """
+    def write(self, value):
+        """Write the value by returning it, instead of storing in a buffer."""
+        return value
+
+
 def render_to_csv_response(queryset, filename=None, append_datestamp=False,
-                           **kwargs):
+                           streaming=True, **kwargs):
     """
     provides the boilerplate for making a CSV http response.
     takes a filename or generates one from the queryset's model.
@@ -42,19 +44,34 @@ def render_to_csv_response(queryset, filename=None, append_datestamp=False,
         filename = generate_filename(queryset,
                                      append_datestamp=append_datestamp)
 
-    response = HttpResponse(content_type='text/csv')
+    response_args = {'content_type': 'text/csv'}
+
+    if not streaming:
+        response = HttpResponse(**response_args)
+        write_csv(queryset, response, **kwargs)
+    else:
+        response = StreamingHttpResponse(
+            _iter_csv(queryset, _Echo(), **kwargs), **response_args)
+
     response['Content-Disposition'] = 'attachment; filename=%s;' % filename
     response['Cache-Control'] = 'no-cache'
 
-    write_csv(queryset, response, **kwargs)
-
     return response
 
 
 def write_csv(queryset, file_obj, **kwargs):
+    """
+    Writes CSV data to a file object based on the contents of the queryset.
+    """
+    # Force iteration over all rows so they all get written to the file
+    for _ in _iter_csv(queryset, file_obj, **kwargs):
+        pass
+
+
+def _iter_csv(queryset, file_obj, **kwargs):
     """
     The main worker function. Writes CSV data to a file object based on the
-    contents of the queryset.
+    contents of the queryset and yields each row.
     """
 
     # process keyword arguments to pull out the ones used by this function
@@ -63,35 +80,47 @@ def write_csv(queryset, file_obj, **kwargs):
     use_verbose_names = kwargs.get('use_verbose_names', True)
     field_order = kwargs.get('field_order', None)
 
-    csv_kwargs = {}
+    csv_kwargs = {'encoding': 'utf-8'}
 
     for key, val in six.iteritems(kwargs):
         if key not in DJQSCSV_KWARGS:
             csv_kwargs[key] = val
 
-    # add BOM to suppor CSVs in MS Excel
-    file_obj.write(u'\ufeff'.encode('utf8'))
+    # add BOM to support CSVs in MS Excel (for Windows only)
+    yield file_obj.write(b'\xef\xbb\xbf')
 
     # the CSV must always be built from a values queryset
     # in order to introspect the necessary fields.
-    if isinstance(queryset, ValuesQuerySet):
+    # However, repeated calls to values can expose fields that were not
+    # present in the original qs. If using `values` as a way to
+    # scope field permissions, this is unacceptable. The solution
+    # is to make sure values is called *once*.
+
+    # perform an string check to avoid a non-existent class in certain
+    # versions
+    if type(queryset).__name__ == 'ValuesQuerySet':
         values_qs = queryset
     else:
-        values_qs = queryset.values()
+        # could be a non-values qs, or could be django 1.9+
+        iterable_class = getattr(queryset, '_iterable_class', object)
+        if iterable_class.__name__ == 'ValuesIterable':
+            values_qs = queryset
+        else:
+            values_qs = queryset.values()
 
     try:
-        field_names = values_qs.field_names
-
+        # Django 1.9+
+        field_names = list(values_qs.query.values_select)
     except AttributeError:
-        # in django1.5, empty querysets trigger
-        # this exception, but not django 1.6
-        raise CSVException("Empty queryset provided to exporter.")
+        # Django 1.8
+        field_names = values_qs.field_names
 
     extra_columns = list(values_qs.query.extra_select)
     if extra_columns:
         field_names += extra_columns
 
-    aggregate_columns = list(values_qs.query.aggregate_select)
+    aggregate_columns = list(values_qs.query.annotation_select)
+
     if aggregate_columns:
         field_names += aggregate_columns
 
@@ -110,20 +139,21 @@ def write_csv(queryset, file_obj, **kwargs):
     name_map = dict((field, field) for field in field_names)
     if use_verbose_names:
         name_map.update(
-            dict((field.name, unicode(field.verbose_name))
+            dict((field.name, field.verbose_name)
                  for field in queryset.model._meta.fields
                  if field.name in field_names))
 
     # merge the custom field headers into the verbose/raw defaults, if provided
     merged_header_map = name_map.copy()
-    merged_header_map.update(field_header_map)
     if extra_columns:
         merged_header_map.update(dict((k, k) for k in extra_columns))
-    writer.writerow(merged_header_map)
+    merged_header_map.update(field_header_map)
+
+    yield writer.writerow(merged_header_map)
 
     for record in values_qs:
-        record = _sanitize_unicode_record(field_serializer_map, record)
-        writer.writerow(record)
+        record = _sanitize_record(field_serializer_map, record)
+        yield writer.writerow(record)
 
 
 def generate_filename(queryset, append_datestamp=False):
@@ -131,7 +161,8 @@ def generate_filename(queryset, append_datestamp=False):
     Takes a queryset and returns a default
     base filename based on the underlying model
     """
-    base_filename = slugify(unicode(queryset.model.__name__)) + '_export.csv'
+    base_filename = slugify(six.text_type(queryset.model.__name__)) \
+        + '_export.csv'
 
     if append_datestamp:
         base_filename = _append_datestamp(base_filename)
@@ -151,11 +182,11 @@ def _validate_and_clean_filename(filename):
         else:
             filename = filename[:-4]
 
-    filename = slugify(unicode(filename)) + '.csv'
+    filename = slugify(six.text_type(filename)) + '.csv'
     return filename
 
 
-def _sanitize_unicode_record(field_serializer_map, record):
+def _sanitize_record(field_serializer_map, record):
 
     def _serialize_value(value):
         # provide default serializer for the case when
@@ -163,23 +194,18 @@ def _sanitize_unicode_record(field_serializer_map, record):
         if isinstance(value, datetime.datetime):
             return value.isoformat()
         else:
-            return unicode(value)
-
-    def _sanitize_text(value):
-        # make sure every text value is of type 'str', coercing unicode
-        if isinstance(value, unicode):
-            return value.encode("utf-8")
-        elif isinstance(value, str):
-            return value
-        else:
-            return str(value).encode("utf-8")
+            return six.text_type(value)
 
     obj = {}
     for key, val in six.iteritems(record):
         if val is not None:
             serializer = field_serializer_map.get(key, _serialize_value)
             newval = serializer(val)
-            obj[_sanitize_text(key)] = _sanitize_text(newval)
+            # If the user provided serializer did not produce a string,
+            # coerce it to a string
+            if not isinstance(newval, six.text_type):
+                newval = six.text_type(newval)
+            obj[key] = newval
 
     return obj
 
